
/*A Table-driven parsing program which reads the file table.dat		*/
/*and then prompts for an input string to be parsed.			*/
/*Entries in table.dat consist of the following fields on each line:	*/
/*	<Symbol (10 characters)> <Tag> <Follow> <Other>			*/
/*<Symbol> = a terminal symbol or the name of a nonterminal;	*/
/*<Tag>    = 0 for terminal nodes, a line number for nonterminal nodes;	*/
/*<Follow> = line number of following node in syntax graph;		*/
/*<Other>  = line number of an alternative node in syntax graph;	*/
/*The <Symbol> field is at least 10 characters wide (including blanks).	*/
/*A <Symbol> consisting of 10 blanks represents the null string.	*/
/*Terminal symbols can be no more than 10 characters long.
/*All input tokens must be separated by blanks.				*/
/*Lines beginning with # are comments and are ignored by the parser.	*/


#include <stdio.h>
#include <string.h>
#define MAX 100
/*Define node for syntax graph */
struct table {
	char symbol[11];
	int  tag;
	int  follow;
	int  other;
};
/*Declare global array of nodes */
struct table  TABLE[MAX];
char sym[11], comment[71];

main()
{
	FILE  *in_file;
	int i;
	char *p;
	int parsegraph();
	void Next();		
				/* read parse table	*/
	if ((in_file=fopen("table.dat","r"))==NULL)
	{ 
		printf("Unable to open syntax file: table.dat\n");
		return(0);
        }
	printf("===========PARSE TABLE===========\n");
	printf("Line  Symbol    Tag  Follow Other\n");
	for(i=1;i<=MAX;)
	{
		if (fscanf(in_file,"%10c",TABLE[i].symbol)==EOF)
			break;
		if (TABLE[i].symbol[0]=='#')	/* read a comment line	*/
			printf("%10s%s",TABLE[i].symbol,
				fgets(comment,70,in_file));
		else
		{	/* terminate symbol field & strip blanks */
			TABLE[i].symbol[10] = 0;
			if (p=strchr(TABLE[i].symbol,' ')) *p = 0;
			/* read tag, follow, other fields	*/
			fscanf(in_file,"%d",&TABLE[i].tag);
			fscanf(in_file,"%d",&TABLE[i].follow);
			fscanf(in_file,"%d",&TABLE[i].other);
			fgets(comment,70,in_file);	/* skip to \n	*/
			printf("%4d  %-10s%3d%6d%6d\n",i,TABLE[i].symbol,
				TABLE[i].tag,TABLE[i].follow,TABLE[i].other);
			i++;
		}
	}
	fclose(in_file);
	printf("Enter input: \n");
	Next();				/* get first token	*/
		/* call parser and check for end of input	*/
	if (parsegraph(1) && (sym[0] == (char)EOF))
		printf(">> valid\n");
	else
		printf(">> invalid\n");	
}

/*Read next input token.  Tokens are separated by whitespace	*/
void Next()
{
	if (scanf("%10s",sym)==EOF)
		sym[0]=EOF;
}

/*Recursive descent table parser	*/
int parsegraph(int top)
{
        int t;
	int flag;
        t=top;
	do
	{
	printf("t = %d, symbol = %s, sym = %s\n",t,TABLE[t].symbol,sym);
		if((TABLE[t].tag) == 0)
		{
			if(strcmp(TABLE[t].symbol, sym)==0)
			{
				flag=1;
				Next();
			}
			else	/* test for empty string */
			{
				flag=(strlen(TABLE[t].symbol)==0);
			}
		}
		else
                	{
			flag=parsegraph(TABLE[t].tag);
		}
		if (flag)
                         t=TABLE[t].follow;
		else
	                 t=TABLE[t].other;
	} while(t);
	printf(" returning flag = %d\n",flag);
	return(flag);
}

